!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Module written by Julien Lesgourgues for using weak lensing data in CosmoMC        !
! This version, written for the COSMOS 3D data by Massey et al. [astro-ph/0701480],  !
! was used in [arXiv:0705.0533] by J. Lesgourgues, M. Viel, M.G. Haehnelt, R. Massey !
! THIS VERSION ASSUMES FLAT UNIVERSE                                                 !
! last change: 9/10/2007                                                             !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

MODULE WeakLen
  use cmbtypes
  use Angular ! ancillary module converting harmonic power spectrum into angular correlation functions
  implicit none

  ! precision parameters:
  integer, parameter :: nlmax=65   ! fixes the maximum required l value in C_l(lensing) 
  ! 65 sufficient for COSMOS data (astro-ph/0701480)
  real, parameter :: dlnl=0.2      ! fixes logarithmic l step for C_l(lensing)
  ! 0.2 small enough for COSMOS data (astro-ph/0701480)  
  real, parameter :: xstop=100.    ! fixes maximum value of x =  l * theta in angular correlation function 
  ! 100 sufficient for COSMOS data (astro-ph/0701480)  
  ! note: integration of C_l J_0,4(l*theta) over l will stop 
  ! at l = min( ll(nlmax),  xstop/theta )

  ! parameters describing the COSMOS data set:
  integer, parameter :: nzmax=151   ! number of discrete z values in selection function files
  integer, parameter :: nbinmax=3   ! number of bins
  integer, parameter :: nthetamax=6 ! number of angles theta

  logical :: Initialize_WeakLen = .false.

  ! arrays (could be replaced by pointers and dynamical allocations...)
  real z(nzmax),etaz(nzmax,nbinmax),eta_norm(nbinmax) ! data's selection functions
  real theta(nthetamax) ! data's theta values
  real d_theo(2*nthetamax*nbinmax) ! For data arranged like a vector
  real d_obs(2*nthetamax*nbinmax)  ! For data arranged like a vector
  real d_sys(2*nthetamax*nbinmax)  ! For data arranged like a vector
  real invcov(2*nthetamax*nbinmax,2*nthetamax*nbinmax) ! Vocariance of data arranged like a vector

CONTAINS

  FUNCTION WeakLenLnLike(CMB, Theory)
    TYPE (CMBParams) CMB
    TYPE (CosmoTheory) Theory
    REAL :: WeakLenLnLike

    integer nl
    real ll(nlmax)
    integer nkmax,nk
    integer zix
    real kk(num_matter_power-1)
    real P(num_matter_power-1,matter_power_lnzsteps)    ! P(k,z)
    real ddP(num_matter_power-1,matter_power_lnzsteps)  ! same splined along k
    real PP(matter_power_lnzsteps)                      ! P(k=l/r, z=z(r))
    real ddPP(matter_power_lnzsteps)                    ! same splined along r
    ! for spline:
    real sig_spline,p_spline,u_spline(num_matter_power-1)
    ! for splint:
    integer k,khi,klo
    real a,b,h
    ! for eta(z):
    integer nz,nbin,nr,nrmax,nrs
    real r(nzmax),dz_o_dr(nzmax),etar(nzmax,nbinmax)
    real g(nzmax,nbinmax)
    character*100 name
    ! for Theory%z_and_r(:,zix)
    real zz(matter_power_lnzsteps),rr(matter_power_lnzsteps),dz_over_dr(matter_power_lnzsteps)
    real ddrr(matter_power_lnzsteps),dddz_over_dr(matter_power_lnzsteps)
    ! For C_l
    real kvalue,PP_r
    real Cl_integrand(nzmax,nbinmax)
    real C(nlmax,nbinmax),ddC(nlmax,nbinmax)
    ! For C(theta)
    real C1(nthetamax,nbinmax),C2(nthetamax,nbinmax)

    if (Feedback >=2) write(*,*)'Enters Weaklen module with:'
    if (Feedback >=2) write(*,*)'matter_power_lnzsteps=',matter_power_lnzsteps,'num_matter_power=',num_matter_power

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Initialization !!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    if (Initialize_Weaklen .eq. .false.) then

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Read data !!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       ! Read eta(z) in cosmos_eta.txt
       name='data/cosmos_eta.txt'
       open (unit=10,file=name)
       if (Feedback >= 2) write(*,*)'Reading file',name
       read(10,*)name
       do nz=1,nzmax
          read(10,*)z(nz),etaz(nz,:)
       end do
       close(10)

       ! Read c(theta) in cosmos_ctheta.txt
       name='data/cosmos_ctheta.txt'
       open (unit=10,file=name)
       if (Feedback >= 2) write(*,*)'Reading file',name
       read(10,*)name
       do k=1,nthetamax
          read(10,*)theta(k),d_obs(k),d_obs(nthetamax+k),d_obs(2*nthetamax+k),&
               d_obs(3*nthetamax+k),d_obs(4*nthetamax+k),d_obs(5*nthetamax+k)
       end do
       close(10)

       ! Read inverse covariance matrix in cosmos_covariance.txt
       name='data/cosmos_invcov.txt'
       open (unit=10,file=name)
       if (Feedback >= 2) write(*,*)'Reading file',name
       ! read first 2 lines of text
       read(10,*)name
       read(10,*)name
       do k=1,2*nthetamax*nbinmax
          read(10,*)invcov(k,:)
       end do
       close(10)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! integrate eta(z) over z (in view of normalizing it to one !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       eta_norm(:)=0.
       do nz=2,nzmax
          eta_norm(:)=eta_norm(:)+0.5*(etaz(nz-1,:)+etaz(nz,:))*(z(nz)-z(nz-1))
       end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Compute Bessel interpolation array !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       call GetBessels(xstop*1.d0) ! subroutine of the "Angular" module

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!! Convert theta values from arcminutes to radians !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       do k=1,nthetamax
          theta(k)=theta(k)/60.*3.14159/180. ! conversion to radians
       end do

       Initialize_Weaklen = .true.

    end if

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! End of initialization !!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!! Various steps for Likelihood computation start here !!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Convert eta(z) to eta(r) !!!!!!!!!!!!!!!!!!!!
!!! (first unnormalized, second normalized to one) !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! array Theory%z_and_r(1,zix) contains list of redshifts zz(zix) in CAMB output 
    ! the first one must be zero:
    if (Theory%z_and_r(1,1) .eq. 0.) then
       write(*,*)'problem: Theory%z_and_r not passed properly. Results zero in Weak lensing module'
       stop
    end if

    ! read array Theory%z_and_r(1,zix) = list of redshifts zz(zix) in CAMB output 
    ! read array Theory%z_and_r(2,zix) = list of corresponding comoving distances rr(zix)
    ! read array Theory%z_and_r(3,zix) = list of corresponding dz/dr (= Hubble parameter)
    do zix=1,matter_power_lnzsteps
       zz(matter_power_lnzsteps+1-zix)=Theory%z_and_r(1,zix)  ! note that z (and r, ...) must be rearranged in growing order for splint
       rr(matter_power_lnzsteps+1-zix)=Theory%z_and_r(2,zix)
       dz_over_dr(matter_power_lnzsteps+1-zix)=Theory%z_and_r(3,zix)
    end do

    ! spline rr and dz/dr as functions of zz: then we can get r and dz/dr at any z
    call spline_real(zz,rr,matter_power_lnzsteps,ddrr)
    call spline_real(zz,dz_over_dr,matter_power_lnzsteps,dddz_over_dr)

    ! for each value of z in the data's selection function file, compute r and dz/dr with splint
    ! compute also the selection function eta(r) = eta(z) dz/dr normalized to one
    ! (inspired from Numerical Recipies's splint)
    nrmax=nzmax
    do nr=1,nrmax
       klo=1
       khi=matter_power_lnzsteps
       do while (khi-klo.gt.1)
          k=(khi+klo)/2
          if(zz(k).gt.z(nr)) then
             khi=k
          else
             klo=k
          endif
       end do
       h=zz(khi)-zz(klo)
       if (h.eq.0.) then
          write(*,*)'Problem: h=0 in splint of rr(z), for nr=',nr,' z(nr)=',z(nr)
          write(*,*)'while klo=',klo,' zz(klo)=',zz(klo),' khi=',khi,' zz(khi)=',zz(khi)
          stop
       end if
       a=(zz(khi)-z(nr))/h
       b=(z(nr)-zz(klo))/h
       r(nr)=a*rr(klo)+b*rr(khi)+((a**3-a)*ddrr(klo)+(b**3-b)*ddrr(khi))*(h**2)/6.
       dz_o_dr(nr)=a*dz_over_dr(klo)+b*dz_over_dr(khi)+((a**3-a)*dddz_over_dr(klo)+(b**3-b)*dddz_over_dr(khi))*(h**2)/6.
       etar(nr,:)=etaz(nr,:)*dz_o_dr(nr)/eta_norm(:)      
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! For each bin, compute window function g(r) = r^2 W(r) a_0/a !!!!!!!!!!!
!!!!!!!! with W(r) = 2 sum_0^r_s_max dr_s eta(r_s) (r_a-r)/r/r_s !!!!!!!!!!!!!!!
!!!!!!!! (simple trapezoidal integration ois enough) !!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    g(1,:)=0.
    do nr=2,nrmax-1
       g(nr,:)=0.
       do nrs=nr+1,nrmax
          g(nr,:)=g(nr,:)+0.5*(etar(nrs,:)*(r(nrs)-r(nr))/r(nrs) &
               +etar(nrs-1,:)*(r(nrs-1)-r(nr))/r(nrs-1))*(r(nrs)-r(nrs-1))
       end do
       g(nr,:)=2.*r(nr)*(1.+z(nr))*g(nr,:)
    end do
    g(nrmax,:)=0.

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Get power spectrum P(k,z) from CAMB !!!!!!!
!!! and prepare its interpolation w.r.t k !!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! fill array of k
    ! Note that we do a shift of 1 in the argument: 
    ! kk(nk) corresponds to k value (unit h/Mpc) for Theory%matter_power(nk+1,zix)
    nkmax=num_matter_power-1
    do nk=1,nkmax
       kk(nk)=exp(log(matter_power_minkh) + nk*matter_power_dlnkh)
    end do
    if (Feedback >= 2) write(*,*)'Weak lensing module: k_max=',kk(nkmax),' h/Mpc =',kk(nkmax)*CMB%h,' 1/Mpc'


    ! fill array of P(k,z)
    ! Note that zz(zix) is arrranged with growing z
    ! this was **already** the case for P(k,zix)
    nkmax=num_matter_power-1
    do zix = 1,matter_power_lnzsteps
       do nk=1,nkmax
          P(nk,zix)=Theory%matter_power(nk+1,zix)
       end do
    end do

    ! for each redshift, spline the P(k,z) w.r.t k
    ! (inspired from Numerical Recipies's spline)
    do zix = 1,matter_power_lnzsteps
       ddP(1,zix)=0.
       u_spline(1)=0.
       do nk=2,nkmax-1
          sig_spline=(kk(nk)-kk(nk-1))/(kk(nk+1)-kk(nk-1))
          p_spline=sig_spline*ddP(nk-1,zix)+2.
          ddP(nk,zix)=(sig_spline-1.)/p_spline
          u_spline(nk)=(6.*((P(nk+1,zix)-P(nk,zix))&
               /(kk(nk+1)-kk(nk))-(P(nk,zix)-P(nk-1,zix))&
               /(kk(nk)-kk(nk-1)))/(kk(nk+1)-kk(nk-1))&
               -sig_spline*u_spline(nk-1))/p_spline
       end do
       ddP(nkmax,zix)=0.
       do nk=nkmax-1,1,-1
          ddP(nk,zix)=ddP(nk,zix)*ddP(nk+1,zix)+u_spline(nk)
       end do
    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! start loop over l  for computation of C_l^shear !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    do nl=1,nlmax

       ! current value of l:
       ll(nl)=1.*exp(dlnl*(nl-1.))

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!! fill array PP(r) = P(l/r,z(r)) !!!!!!!!!!!!!!
!!! prepare it for interpolation along r  !!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       PP(1)=0.
       do zix = 2,matter_power_lnzsteps
          kvalue=ll(nl)/rr(zix)/CMB%h   ! k=l/r in units h/Mpc
          if ((kvalue .le. kk(1)) .or. (kvalue .ge. kk(nkmax))) then
             PP(zix)=0.
          else
             ! compute P(k,z) at k=l/r
             ! (inspired from Numerical Recipies's splint)
             klo=1
             khi=nkmax
             do while (khi-klo.gt.1)
                k=(khi+klo)/2
                if(kk(k).gt.kvalue) then
                   khi=k
                else
                   klo=k
                endif
             end do
             h=kk(khi)-kk(klo)
             if (h.eq.0.) then
                write(*,*)'Problem: h=0 in splint of P(k,z)'
                stop
             end if
             a=(kk(khi)-kvalue)/h
             b=(kvalue-kk(klo))/h
             PP(zix)=a*P(klo,zix)+b*P(khi,zix)+((a**3-a)*ddP(klo,zix)+(b**3-b)*ddP(khi,zix))*(h**2)/6.
          end if
       end do
       call spline_real(rr,PP,matter_power_lnzsteps,ddPP)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Find PP(r) = P(l/r,z(r)) for each r(nr) !!!!!!!!!!!!!!!!
!!! and infer Cl_integrand = (g(r) / r)**2 * P(l/r,z(r)) !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       Cl_integrand(1,:)=0.
       do nr=2,nrmax
          ! (inspired from Numerical Recipies's splint)
          klo=1
          khi=matter_power_lnzsteps
          do while (khi-klo.gt.1)
             k=(khi+klo)/2
             if(rr(k).gt.r(nr)) then
                khi=k
             else
                klo=k
             endif
          end do
          h=rr(khi)-rr(klo)
          if (h.eq.0.) then
             write(*,*)'Problem: h=0 in splint of rr(z)'
             stop
          end if
          a=(rr(khi)-r(nr))/h
          b=(r(nr)-rr(klo))/h
          PP_r=a*PP(klo)+b*PP(khi)+((a**3-a)*ddPP(klo)+(b**3-b)*ddPP(khi))*(h**2)/6.
          PP_r=PP_r/(CMB%h)**3  ! in units Mpc**3
          Cl_integrand(nr,:)=(g(nr,:)/r(nr))**2*PP_r
       end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Integrate over r to get C_l^shear !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! C_l^shear = 9/16 Omega_m0^2 H_0^4 \sum_0^rmax dr (g(r)/r)^2 P(kl/r,r) !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       C(nl,:)=0.
       do nr=2,nrmax
          C(nl,:)=C(nl,:)+0.5*(Cl_integrand(nr,:)+Cl_integrand(nr-1,:))*(r(nr)-r(nr-1)) 
       end do
       C(nl,:)= C(nl,:)*9./16.*(CMB%omb+CMB%omdm)**2 ! in units Mpc**4
       C(nl,:)= C(nl,:)*(CMB%h/2997.9)**4 ! dimensionless

    end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!! Spline Cl(nl,nbin) along l !!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    do nbin = 1,nbinmax
       ! (inspired from Numerical Recipies's spline)
       ddC(1,nbin)=0.
       u_spline(1)=0.
       do nl=2,nlmax-1
          sig_spline=(ll(nl)-ll(nl-1))/(ll(nl+1)-ll(nl-1))
          p_spline=sig_spline*ddC(nl-1,nbin)+2.
          ddC(nl,nbin)=(sig_spline-1.)/p_spline
          u_spline(nl)=(6.*((C(nl+1,nbin)-C(nl,nbin))&
               /(ll(nl+1)-ll(nl))-(C(nl,nbin)-C(nl-1,nbin))&
               /(ll(nl)-ll(nl-1)))/(ll(nl+1)-ll(nl-1))&
               -sig_spline*u_spline(nl-1))/p_spline
       end do
       ddC(nlmax,nbin)=0.
       do nl=nlmax-1,1,-1
          ddC(nl,nbin)=ddC(nl,nbin)*ddC(nl+1,nbin)+u_spline(nl)
       end do
    end do

    if (Feedback >=2) write(*,*)'lmax =',ll(nlmax)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!! Compute correlation function C_1,2(theta) (using module "Angular") !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    call Cl_to_Ctheta(ll,C,ddC,theta,c1,c2,nlmax,nthetamax,nbinmax,xstop)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Arrange all C(theta) values in a vector, following the same order as in the data !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    nl=0
    do nbin=1,nbinmax
       do nr=1,2
          do k=1,nthetamax
             nl=nl+1
             if (nr==1) then
                d_theo(nl)=c1(k,nbin)
             else
                d_theo(nl)=c2(k,nbin)
             end if
          end do
       end do
    end do

!!$    do k=1,nthetamax
!!$          write(9,*)theta(k),d_theo(k-1+1)+d_theo(k-1+7),d_obs(k-1+1)+d_obs(k-1+7)
!!$          write(9,*)theta(k),d_theo(k-1+13)+d_theo(k-1+19),d_obs(k-1+13)+d_obs(k-1+19)
!!$          write(9,*)theta(k),d_theo(k-1+25)+d_theo(k-1+31),d_obs(k-1+25)+d_obs(k-1+31)
!!$    end do

!!$    do k=1,36
!!$       write(9,*)d_theo(k),d_obs(k)
!!$    end do

    ! compute the "rescaling vector" accounting for systematic errors,
    ! described by three nuisance parameters [arXiv:0705.0533]
    ! A is gaussian distributed with average 1 and sigma_A=0.06
    ! B is half-gaussian distributed with B>1 and sigma_B=0.05
    ! C is half-gaussian distributed with C>1 and sigma_C=0.10
    ! first bin (low z)
    do nl=1,12
       d_sys(nl)=(CMB%norm(norm_A)/CMB%norm(norm_B))**2
    end do
    ! second bin (mid z)
    do nl=13,24
       d_sys(nl)=(CMB%norm(norm_A))**2
    end do
    ! third bin (high z)
    do nl=25,36
       d_sys(nl)=(CMB%norm(norm_A)*CMB%norm(norm_B))**2/CMB%norm(norm_C)
    end do
    ! compute chi2=-2lnL corresponding to priors on A,B,C (modulo a constant)    
    WeakLenLnLike=((CMB%norm(norm_A)-1)/0.06)**2+((CMB%norm(norm_B)-1)/0.05)**2+((CMB%norm(norm_C)-1)/0.10)**2
    if (Feedback >=2) write(*,*)'WL nuisnace parameters A,B,C =',CMB%norm(norm_A),CMB%norm(norm_B),CMB%norm(norm_C)

    ! add the chi2
    do k=1,2*nthetamax*nbinmax
       do nk=1,2*nthetamax*nbinmax
          WeakLenLnLike=WeakLenLnLike+&
               (d_theo(k)-d_sys(k)*d_obs(k))*&
               (invcov(k,nk)/d_sys(k)/d_sys(nk))*&
               (d_theo(nk)-d_sys(nk)*d_obs(nk))

       end do
    end do

    ! compute -lnL = chi2/2
    WeakLenLnLike=WeakLenLnLike/2.

    if (Feedback >=2) write(*,*)'Weak lensing -LnLike=',WeakLenLnLike

  END FUNCTION WeakLenLnLike

end module WeakLen

